#!/usr/bin/env python3
from contextlib import contextmanager


class GraphDatabase:

    def __init__(self, storage):
        self._storage = storage
        self._indices = list()

    def close(self):
        self._storage.close()

    def index(self, name):
        """Instruct the database to index `name` properties

        Every property `name` that is a `str` with less than 132 `char`
        or an `int`"""
        self._indices.append(name)

    def txn(self, sync=True, *args, **kwargs):
        """Return a storage specific transaction.

        See each storage for further informations regarding the keywords
        arguments. Here is a brief overview what can be supported:

        ``sync`` must be true, if you want transactions to be immediatly sync
        to disk changes made to the database. Setting ``sync=False`` will
        provide a speed boost you might loose data, but not corrupt
        your database.
        """
        return self._storage._txn(self._indices, sync)

    @contextmanager
    def transaction(self, *args, **kwargs):
        """Context manager to start a new transaction.

        See ``GraphDatabase.txn()`` to know the meaning of ``kwargs``"""
        txn = self.txn(*args, **kwargs)
        try:
            yield txn
        except Exception as exc:
            txn.rollback()
            raise exc
        else:
            txn.commit()
